--[[
	TODO:
		- Maybe add tooltip to indicate whenever double-clicking is going to do something fancy?
		- Maybe use OnDoubleClick for buttons (which I didn't know about before)... Still need to do manual OnClick for non-buttons, so maybe not worth the effort
]]

local addonName = ...

local function FindFirstCompletedQuest()
	for i = 1, GetNumActiveQuests() do
		if select(2, GetActiveTitle(i)) then
			return i
		end
	end
	return 1 -- Fall back to selecting first incompleted quest if none are completed
end

local function FindFirstNonRepeatableGossipQuest(...) -- GetGossipAvailableQuests returns title#, level#, isLowLevel#, isDaily#, isRepeatable * GetNumGossipAvailableQuests()
	for i = 1, select("#", ...), 5 do
		local daily, repeatable = select(i + 3, ...)
		if not daily and not repeatable then
			return ceil(i / 5)
		end
	end
end

local function FindFirstCompletedGossipQuest(...) -- GetGossipActiveQuests returns title#, level#, isLowLevel#, isCompleted# * GetNumGossipActiveQuests()
	for i = 1, select("#", ...), 4 do
		if select(i + 3, ...) then -- Check if isCompleted is true for the quest
			return ceil(i / 4)
		end
	end
end

local function FindFirstNonbinderGossipOption(...) -- Favor non-binder (set-my-hearthstone) gossips, if any are available
	for i = 1, select("#", ...), 2 do
		if select(i + 1, ...) ~= "binder" then
			return ceil(i / 2)
		end
	end
	return 1
end

 -- The functions that get called when the frames are double-clicked
local frameHandlers = {
	[GossipFrame] = function(button) -- This seems to show up a lot more than QuestFrameGreetingPanel, which seems a bit odd
		if button == "RightButton" then
			GossipFrameGreetingGoodbyeButton:GetScript("OnClick")()
		else
			local availableIndex = FindFirstNonRepeatableGossipQuest(GetGossipAvailableQuests())
			local completedIndex = FindFirstCompletedGossipQuest(GetGossipActiveQuests())
			if availableIndex then -- There's a non-repeatable available quest
				SelectGossipAvailableQuest(availableIndex)
			elseif completedIndex then -- There's an active completed quest
				SelectGossipActiveQuest(completedIndex)
			elseif GetNumGossipAvailableQuests() > 0 then -- No non-repeatable active or completed quests, fall back to repeatables
				SelectGossipAvailableQuest(1)
			elseif GetNumGossipOptions() > 0 then -- No active or completed quests, but there are gossip options. Prefer first non-binder gossip
				SelectGossipOption(FindFirstNonbinderGossipOption(GetGossipOptions()))
			elseif GetNumGossipActiveQuests() > 0 then -- Fall back to first non-completed normal quest
				SelectGossipActiveQuest(1)
			end
		end
	end,
	[QuestFrameGreetingPanel] = function(button) -- This seems to be used for questgivers that have repeatable quests. Actually scratch that, I just found a questgiver that opens gossipframe, and has repeatable quest. What a mess
		if button == "RightButton" then
			QuestFrameGreetingGoodbyeButton:GetScript("OnClick")()
		elseif GetNumAvailableQuests() > 0 then
			SelectAvailableQuest(1) -- Can these indices ever start at >= 2? And shlould I check for repeatables here too?
		elseif GetNumActiveQuests() > 0 then
			SelectActiveQuest(FindFirstCompletedQuest())
		end
	end,
	[QuestLogFrame] = function(button) -- Should this also let you track/untrack quests? I personally don't use it much, but could do it with modifier+doubleclick I guess
		if button == "RightButton" then
			local dialog = StaticPopup_FindVisible("ABANDON_QUEST") or StaticPopup_FindVisible("ABANDON_QUEST_WITH_ITEMS")
			if dialog then
				return StaticPopup_OnClick(dialog, 1)
			end
			QuestLogFrameAbandonButton:GetScript("OnClick")()
		else
			QuestLogPushQuest()
		end
	end,
	[QuestFrameDetailPanel] = QuestDetailAcceptButton_OnClick, -- These three get the hide-on-rightclick later on in the OnClick function
	[QuestFrameProgressPanel] = QuestProgressCompleteButton_OnClick,
	[QuestFrameRewardPanel] = QuestRewardCompleteButton_OnClick,
}
frameHandlers[QuestLogDetailFrame] = frameHandlers[QuestLogFrame]

-- Our custom OnClick handler
local lastTargetFrame
local lastClickTime = 0
local lastButton = ""
local function OnClick(self, button)
	if not frameHandlers[self] then -- Loop through the frameHandlers table if the frame isn't in it to see if any of the frames are visible, to figure out what we're supposed to be doing
		for frame in pairs(frameHandlers) do
			if frame:IsVisible() then
				return OnClick(frame, button)
			end
		end
	end
	local now = GetTime()
	if now - lastClickTime < 0.5 and self:IsShown() and lastTargetFrame == self and lastButton == button then
		if button == "RightButton" and (self == QuestFrameDetailPanel or self == QuestFrameProgressPanel or self == QuestFrameRewardPanel) then
			QuestGoodbyeButton_OnClick()
		else
			frameHandlers[self](button)
		end
		lastClickTime, lastTargetFrame, lastButton = 0, self, button
		return
	end
	lastClickTime, lastTargetFrame, lastButton = now, self, button
end

-- Hook the frame and perform common tasks
local function HookFrame(name, scriptType, handler)
	local frame = type(name) == "string" and _G[name] or name
	frame:HookScript(scriptType, handler)
	if frame.RegisterForClicks then
		frame:RegisterForClicks("LeftButtonUp", "RightButtonUp") -- Make sure it's getting both left and right clicks
	end
	frame:EnableMouse()
	return frame
end

-- Piggyback above to hook and put the frame in to frameHandlers
local function HookFrameUsingOnClick(name, handler)
	frameHandlers[HookFrame(name, "OnClick", OnClick)] = handler -- Ok now I'm just code-golfing >_>
end

-- First lets hook the frames we got in frameHandlers
for frame in pairs(frameHandlers) do
	HookFrame(frame, "OnMouseUp", OnClick)
end

-- These buttons aren't added to the frameHandlers table to save memory
for i = 1, 6 do
	HookFrame("QuestProgressItem"..i, "OnClick", OnClick)
end

 -- These buttons are shared between the QuestFrameDetailPanel and QuestFrameRewardPanel, and we don't want the reward type to be double-clickable so we don't accidentally select the wrong rewards when clicking really fast.
local function QuestInfoItemHandler(self, button)
	if button == "RightButton" or not QuestFrameRewardPanel:IsVisible() then
		OnClick(self, button)
	end
end
for i = 1, 10 do
	HookFrame("QuestInfoItem"..i, "OnClick", QuestInfoItemHandler)
end

-- Accept the annoying popups confirming rolls
local function LootFrameHandler()
	local dialog = StaticPopup_FindVisible("CONFIRM_LOOT_ROLL")
	if dialog then
		StaticPopup_OnClick(dialog, 1)
	end
end
for i = 1, 4 do
	HookFrameUsingOnClick("GroupLootFrame"..i.."RollButton", LootFrameHandler)
	HookFrameUsingOnClick("GroupLootFrame"..i.."GreedButton", LootFrameHandler)
	HookFrameUsingOnClick("GroupLootFrame"..i.."DisenchantButton", LootFrameHandler)
end

HookFrameUsingOnClick = nil -- Don't need this anymore

----------------------------------------------------
-- Tradeskill list button stuff

local f = CreateFrame("frame")
f:Hide()
f:RegisterEvent("ADDON_LOADED")

-- Delay loading by one frame, so addons modifying the tradeskill frame (such as FramesResized) fire first
f:SetScript("OnUpdate", function(self)
	local function TradeSkillHandler()
		TradeSkillCreateButton:GetScript("OnClick")()
	end
	for i = 1, TRADE_SKILLS_DISPLAYED do
		HookFrame("TradeSkillSkill"..i, "OnDoubleClick", TradeSkillHandler) -- Uses OnDoubleClick so don't need to worry about our own OnClick solution.
	end

	self:UnregisterAllEvents()
	self:SetScript("OnEvent", nil)
	self:SetScript("OnUpdate", nil)
	HookFrame = nil -- Don't need this anymore
end)

f:SetScript("OnEvent", function(self, event, addon)
	if addon == "Blizzard_TradeSkillUI" then
		self:Show()
	end
end)

----------------------------------------------------

-- Slashcommand to show version info, 'cos having to open the TOC to check it sucks
_G["SLASH_"..addonName:upper().."1"], _G["SLASH_"..addonName:upper().."2"] = "/"..addonName:lower(), "/"..addonName:lower().."version"
SlashCmdList[addonName:upper()] = function()
	print(addonName, "version:", GetAddOnMetadata(addonName, "Version"))
end
